# import json
# import pytest

# from langchain_openai import ChatOpenAI
# from lmnr.traceloop_sdk.decorators import task
# from opentelemetry.semconv_ai import SpanAttributes


# @pytest.mark.vcr
# def test_task_io_serialization_with_langchain(exporter):
#     @task(name="answer_question")
#     def answer_question():
#         chat = ChatOpenAI(temperature=0)

#         return chat.invoke("Is Berlin the capital of Germany? Answer with yes or no")

#     answer_question()

#     spans = exporter.get_finished_spans()

#     assert [span.name for span in spans] == [
#         "ChatOpenAI.chat",
#         "answer_question.task",
#     ]

#     task_span = next(span for span in spans if span.name == "answer_question.task")
#     assert (
#         json.loads(task_span.attributes.get(SpanAttributes.TRACELOOP_ENTITY_OUTPUT))[
#             "kwargs"
#         ]["content"]
#         == "Yes"
#     )



import datetime
from enum import Enum
import pydantic
import requests
from typing import Any, Awaitable, Callable, Literal, Optional, Union
import uuid

from .utils import serialize


class ChatMessage(pydantic.BaseModel):
    role: str
    content: str


class ConditionedValue(pydantic.BaseModel):
    condition: str
    value: "NodeInput"


Numeric = Union[int, float]
NumericTypes = (int, float)  # for use with isinstance

NodeInput = Union[str, list[ChatMessage], ConditionedValue, Numeric, bool]
PipelineOutput = Union[NodeInput]


class PipelineRunRequest(pydantic.BaseModel):
    inputs: dict[str, NodeInput]
    pipeline: str
    env: dict[str, str] = pydantic.Field(default_factory=dict)
    metadata: dict[str, str] = pydantic.Field(default_factory=dict)
    stream: bool = pydantic.Field(default=False)
    parent_span_id: Optional[uuid.UUID] = pydantic.Field(default=None)
    trace_id: Optional[uuid.UUID] = pydantic.Field(default=None)

    # uuid is not serializable by default, so we need to convert it to a string
    def to_dict(self):
        return {
            "inputs": {
                k: v.model_dump() if isinstance(v, pydantic.BaseModel) else serialize(v)
                for k, v in self.inputs.items()
            },
            "pipeline": self.pipeline,
            "env": self.env,
            "metadata": self.metadata,
            "stream": self.stream,
            "parentSpanId": str(self.parent_span_id) if self.parent_span_id else None,
            "traceId": str(self.trace_id) if self.trace_id else None,
        }


class PipelineRunResponse(pydantic.BaseModel):
    outputs: dict[str, dict[str, PipelineOutput]]
    run_id: str


class PipelineRunError(Exception):
    error_code: str
    error_message: str

    def __init__(self, response: requests.Response):
        try:
            resp_json = response.json()
            self.error_code = resp_json["error_code"]
            self.error_message = resp_json["error_message"]
            super().__init__(self.error_message)
        except Exception:
            super().__init__(response.text)

    def __str__(self) -> str:
        try:
            return str(
                {"error_code": self.error_code, "error_message": self.error_message}
            )
        except Exception:
            return super().__str__()


EvaluationDatapointData = dict[str, Any]
EvaluationDatapointTarget = dict[str, Any]


# EvaluationDatapoint is a single data point in the evaluation
class Datapoint(pydantic.BaseModel):
    # input to the executor function. Must be a dict with string keys
    data: EvaluationDatapointData
    # input to the evaluator function (alongside the executor output).
    # Must be a dict with string keys
    target: EvaluationDatapointTarget


ExecutorFunctionReturnType = Any
EvaluatorFunctionReturnType = Union[Numeric, dict[str, Numeric]]

ExecutorFunction = Callable[
    [EvaluationDatapointData, Any, dict[str, Any]],
    Union[ExecutorFunctionReturnType, Awaitable[ExecutorFunctionReturnType]],
]

# EvaluatorFunction is a function that takes the output of the executor and the
# target data, and returns a score. The score can be a single number or a
# record of string keys and number values. The latter is useful for evaluating
# multiple criteria in one go instead of running multiple evaluators.
EvaluatorFunction = Callable[
    [ExecutorFunctionReturnType, Any, dict[str, Any]],
    Union[EvaluatorFunctionReturnType, Awaitable[EvaluatorFunctionReturnType]],
]

EvaluationStatus = Literal["Started", "Finished", "Error"]


class CreateEvaluationResponse(pydantic.BaseModel):
    id: uuid.UUID
    createdAt: datetime.datetime
    name: str
    status: EvaluationStatus
    projectId: uuid.UUID
    metadata: Optional[dict[str, Any]] = None
    averageScores: Optional[dict[str, Numeric]] = None


UpdateEvaluationResponse = CreateEvaluationResponse


class EvaluationResultDatapoint(pydantic.BaseModel):
    data: EvaluationDatapointData
    target: EvaluationDatapointTarget
    executor_output: ExecutorFunctionReturnType
    scores: dict[str, Numeric]
    trace_id: uuid.UUID

    # uuid is not serializable by default, so we need to convert it to a string
    def to_dict(self):
        return {
            "data": {
                k: v.model_dump() if isinstance(v, pydantic.BaseModel) else serialize(v)
                for k, v in self.data.items()
            },
            "target": {
                k: v.model_dump() if isinstance(v, pydantic.BaseModel) else serialize(v)
                for k, v in self.target.items()
            },
            "executorOutput": serialize(self.executor_output),
            "scores": self.scores,
            "traceId": str(self.trace_id),
        }


class SpanType(Enum):
    DEFAULT = "DEFAULT"
    LLM = "LLM"
    PIPELINE = "PIPELINE"  # must not be set manually
    EXECUTOR = "EXECUTOR"
    EVALUATOR = "EVALUATOR"
    EVALUATION = "EVALUATION"


class TraceType(Enum):
    DEFAULT = "DEFAULT"
    EVENT = "EVENT"  # must not be set manually
    EVALUATION = "EVALUATION"



from lmnr.traceloop_sdk.instruments import Instruments
from opentelemetry import context
from opentelemetry.trace import (
    INVALID_SPAN,
    get_current_span,
    SpanKind,
)
from opentelemetry.util.types import AttributeValue
from opentelemetry.context.context import Context
from opentelemetry.util import types
from lmnr.traceloop_sdk import Traceloop
from lmnr.traceloop_sdk.tracing import get_tracer
from contextlib import contextmanager
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from pydantic.alias_generators import to_snake
from typing import Any, Optional, Set, Union

import copy
import datetime
import dotenv
import json
import logging
import os
import requests
import uuid

from lmnr.traceloop_sdk.tracing.attributes import (
    SESSION_ID,
    SPAN_INPUT,
    SPAN_OUTPUT,
    TRACE_TYPE,
    USER_ID,
)
from lmnr.traceloop_sdk.tracing.tracing import (
    set_association_properties,
    update_association_properties,
)

from .log import VerboseColorfulFormatter

from .types import (
    CreateEvaluationResponse,
    EvaluationResultDatapoint,
    PipelineRunError,
    PipelineRunResponse,
    NodeInput,
    PipelineRunRequest,
    TraceType,
    UpdateEvaluationResponse,
)


class Laminar:
    __base_http_url: str
    __base_grpc_url: str
    __project_api_key: Optional[str] = None
    __env: dict[str, str] = {}
    __initialized: bool = False

    @classmethod
    def initialize(
        cls,
        project_api_key: Optional[str] = None,
        env: dict[str, str] = {},
        base_url: Optional[str] = None,
        http_port: Optional[int] = None,
        grpc_port: Optional[int] = None,
        instruments: Optional[Set[Instruments]] = None,
    ):
        """Initialize Laminar context across the application.
        This method must be called before using any other Laminar methods or
        decorators.

        Args:
            project_api_key (Optional[str], optional): Laminar project api key.
                            You can generate one by going to the projects
                            settings page on the Laminar dashboard.
                            If not specified, it will try to read from the
                            LMNR_PROJECT_API_KEY environment variable
                            in os.environ or in .env file.
                            Defaults to None.
            env (dict[str, str], optional): Default environment passed to
                            `run` and `evaluate_event` requests, unless
                            overriden at request time. Usually, model
                            provider keys are stored here.
                            Defaults to {}.
            base_url (Optional[str], optional): Laminar API url.
                            If not specified, defaults to https://api.lmnr.ai.
            http_port (Optional[int], optional): Laminar API http port.
                            If not specified, defaults to 443.
            grpc_port (Optional[int], optional): Laminar API grpc port.
                            If not specified, defaults to 8443.

        Raises:
            ValueError: If project API key is not set
        """
        cls.__project_api_key = project_api_key or os.environ.get(
            "LMNR_PROJECT_API_KEY"
        )
        if not cls.__project_api_key:
            dotenv_path = dotenv.find_dotenv(usecwd=True)
            cls.__project_api_key = dotenv.get_key(
                dotenv_path=dotenv_path, key_to_get="LMNR_PROJECT_API_KEY"
            )
        if not cls.__project_api_key:
            raise ValueError(
                "Please initialize the Laminar object with"
                " your project API key or set the LMNR_PROJECT_API_KEY"
                " environment variable in your environment or .env file"
            )

        cls.__base_http_url = f"{base_url or 'https://api.lmnr.ai'}:{http_port or 443}"
        cls.__base_grpc_url = f"{base_url or 'https://api.lmnr.ai'}:{grpc_port or 8443}"

        cls.__env = env
        cls.__initialized = True
        cls._initialize_logger()
        Traceloop.init(
            exporter=OTLPSpanExporter(
                endpoint=cls.__base_grpc_url,
                headers={"authorization": f"Bearer {cls.__project_api_key}"},
            ),
            instruments=instruments,
        )

    @classmethod
    def is_initialized(cls):
        """Check if Laminar is initialized. A utility to make sure other
        methods are called after initialization.

        Returns:
            bool: True if Laminar is initialized, False otherwise
        """
        return cls.__initialized

    @classmethod
    def _initialize_logger(cls):
        cls.__logger = logging.getLogger(__name__)
        console_log_handler = logging.StreamHandler()
        console_log_handler.setFormatter(VerboseColorfulFormatter())
        cls.__logger.addHandler(console_log_handler)

    @classmethod
    def run(
        cls,
        pipeline: str,
        inputs: dict[str, NodeInput],
        env: dict[str, str] = {},
        metadata: dict[str, str] = {},
        parent_span_id: Optional[uuid.UUID] = None,
        trace_id: Optional[uuid.UUID] = None,
    ) -> PipelineRunResponse:
        """Runs the pipeline with the given inputs

        Args:
            pipeline (str): name of the Laminar pipeline.
                The pipeline must have a target version set.
            inputs (dict[str, NodeInput]):
                inputs to the endpoint's target pipeline.
                Keys in the dictionary must match input node names
            env (dict[str, str], optional):
                Environment variables for the pipeline execution.
                Defaults to {}.
            metadata (dict[str, str], optional):
                any custom metadata to be stored
                with execution trace. Defaults to {}.
            parent_span_id (Optional[uuid.UUID], optional):
                parent span id for the resulting span.
                Defaults to None.
            trace_id (Optional[uuid.UUID], optional):
                trace id for the resulting trace.
                Defaults to None.

        Returns:
            PipelineRunResponse: response object containing the outputs

        Raises:
            ValueError: if project API key is not set
            PipelineRunError: if the endpoint run fails
        """
        if cls.__project_api_key is None:
            raise ValueError(
                "Please initialize the Laminar object with your project "
                "API key or set the LMNR_PROJECT_API_KEY environment variable"
            )
        try:
            current_span = get_current_span()
            if current_span != INVALID_SPAN:
                parent_span_id = parent_span_id or uuid.UUID(
                    int=current_span.get_span_context().span_id
                )
                trace_id = trace_id or uuid.UUID(
                    int=current_span.get_span_context().trace_id
                )
            request = PipelineRunRequest(
                inputs=inputs,
                pipeline=pipeline,
                env=env,
                metadata=metadata,
                parent_span_id=parent_span_id,
                trace_id=trace_id,
            )
        except Exception as e:
            raise ValueError(f"Invalid request: {e}")

        response = requests.post(
            cls.__base_http_url + "/v1/pipeline/run",
            data=json.dumps(request.to_dict()),
            headers=cls._headers(),
        )
        if response.status_code != 200:
            raise PipelineRunError(response)
        try:
            resp_json = response.json()
            keys = list(resp_json.keys())
            for key in keys:
                value = resp_json[key]
                del resp_json[key]
                resp_json[to_snake(key)] = value
            return PipelineRunResponse(**resp_json)
        except Exception:
            raise PipelineRunError(response)

    @classmethod
    def event(
        cls,
        name: str,
        value: Optional[AttributeValue] = None,
        timestamp: Optional[Union[datetime.datetime, int]] = None,
    ):
        """Associate an event with the current span. If event with such name never
        existed, Laminar will create a new event and infer its type from the value.
        If the event already exists, Laminar will append the value to the event
        if and only if the value is of a matching type. Otherwise, the event won't
        be recorded Supported types are string, numeric, and boolean. If the value
        is `None`, event is considered a boolean tag with the value of `True`.

        Args:
            name (str): event name
            value (Optional[AttributeValue]): event value. Must be a primitive type.
                            Boolean true is assumed in the backend if value is None.
                            Defaults to None.
            timestamp (Optional[Union[datetime.datetime, int]], optional):
                            If int, must be epoch nanoseconds. If not
                            specified, relies on the underlying OpenTelemetry
                            implementation. Defaults to None.
        """
        if timestamp and isinstance(timestamp, datetime.datetime):
            timestamp = int(timestamp.timestamp() * 1e9)

        event = {
            "lmnr.event.type": "default",
        }
        if value is not None:
            event["lmnr.event.value"] = value

        current_span = get_current_span()
        if current_span == INVALID_SPAN:
            cls.__logger.warning(
                "`Laminar().event()` called outside of span context. "
                f"Event '{name}' will not be recorded in the trace. "
                "Make sure to annotate the function with a decorator"
            )
            return

        current_span.add_event(name, event, timestamp)

    @classmethod
    def evaluate_event(
        cls,
        name: str,
        evaluator: str,
        data: dict[str, AttributeValue],
        env: Optional[dict[str, str]] = None,
        timestamp: Optional[Union[datetime.datetime, int]] = None,
    ):
        """Send an event for evaluation to the Laminar backend

        Args:
            name (str): name of the event
            evaluator (str): name of the pipeline that evaluates the event.
                        The pipeline must have a target version set.
            data (dict[str, AttributeValue]): map from input node name to
                        its value in the evaluator pipeline
            env (dict[str, str], optional): environment variables required
                        to run the pipeline. Defaults to {}.
            timestamp (Optional[Union[datetime.datetime, int]], optional):
                        If int, must be epoch nanoseconds.
                        If not specified, relies on the underlying
                        OpenTelemetry implementation. Defaults to None.
        """
        if timestamp and isinstance(timestamp, datetime.datetime):
            timestamp = int(timestamp.timestamp() * 1e9)
        event = {
            "lmnr.event.type": "evaluate",
            "lmnr.event.evaluator": evaluator,
            "lmnr.event.data": json.dumps(data),
            "lmnr.event.env": json.dumps(env if env is not None else cls.__env),
        }
        current_span = get_current_span()
        if current_span == INVALID_SPAN:
            cls.__logger.warning(
                "`Laminar().evaluate_event()` called outside of span context."
                f"Event '{name}' will not be recorded in the trace. "
                "Make sure to annotate the function with a decorator"
            )
            return

        current_span.add_event(name, event, timestamp)

    @classmethod
    @contextmanager
    def start_as_current_span(
        cls,
        name: str,
        input: Any = None,
        context: Optional[Context] = None,
        kind: SpanKind = SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links=None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        end_on_exit: bool = True,
    ):
        """Start a new span as the current span. Useful for manual instrumentation.
        This is the preferred and more stable way to use manual instrumentation.

        Usage example:
        ```python
        with Laminar.start_as_current_span("my_span", input="my_input"):
            await my_async_function()
        ```

        Args:
            name (str): name of the span
            input (Any, optional): input to the span. Will be sent as an
                attribute, so must be json serializable. Defaults to None.
            context (Optional[Context], optional): context to start the span in.
                Defaults to None.
            kind (SpanKind, optional): kind of the span. Defaults to SpanKind.INTERNAL.
            attributes (types.Attributes, optional): attributes to set on the span.
                Defaults to None.
            links ([type], optional): links to set on the span. Defaults to None.
            start_time (Optional[int], optional): start time of the span.
                Defaults to None.
            record_exception (bool, optional): whether to record exceptions.
                Defaults to True.
            set_status_on_exception (bool, optional): whether to set status on exception.
                Defaults to True.
            end_on_exit (bool, optional): whether to end the span on exit.
                Defaults to True.
        """
        with get_tracer() as tracer:
            with tracer.start_as_current_span(
                name,
                context=context,
                kind=kind,
                attributes=attributes,
                links=links,
                start_time=start_time,
                record_exception=record_exception,
                set_status_on_exception=set_status_on_exception,
                end_on_exit=end_on_exit,
            ) as span:
                if input is not None:
                    span.set_attribute(
                        SPAN_INPUT,
                        json.dumps(input),
                    )
                yield span

    @classmethod
    def set_span_output(cls, output: Any = None):
        """Set the output of the current span. Useful for manual instrumentation.

        Args:
            output (Any, optional): output of the span. Will be sent as an
                attribute, so must be json serializable. Defaults to None.
        """
        span = get_current_span()
        if output is not None and span != INVALID_SPAN:
            span.set_attribute(SPAN_OUTPUT, json.dumps(output))

    @classmethod
    def set_session(
        cls,
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ):
        """Set the session and user id for the current span and the context
        (i.e. any children spans created from the current span in the current
        thread).

        Args:
            session_id (Optional[str], optional): Custom session id.
                            Useful to debug and group long-running
                            sessions/conversations.
                            Defaults to None.
            user_id (Optional[str], optional): Custom user id.
                            Useful for grouping spans or traces by user.
                            Defaults to None.
        """
        association_properties = {}
        if session_id is not None:
            association_properties[SESSION_ID] = session_id
        if user_id is not None:
            association_properties[USER_ID] = user_id
        update_association_properties(association_properties)

    @classmethod
    def _set_trace_type(
        cls,
        trace_type: TraceType,
    ):
        """Set the trace_type for the current span and the context
        Args:
            trace_type (TraceType): Type of the trace
        """
        association_properties = {
            TRACE_TYPE: trace_type.value,
        }
        update_association_properties(association_properties)

    @classmethod
    def clear_session(cls):
        """Clear the session and user id from  the context"""
        props: dict = copy.copy(context.get_value("association_properties"))
        props.pop("session_id", None)
        props.pop("user_id", None)
        set_association_properties(props)

    @classmethod
    def create_evaluation(cls, name: Optional[str]) -> CreateEvaluationResponse:
        response = requests.post(
            cls.__base_http_url + "/v1/evaluations",
            data=json.dumps({"name": name}),
            headers=cls._headers(),
        )
        if response.status_code != 200:
            try:
                resp_json = response.json()
                raise ValueError(f"Error creating evaluation {json.dumps(resp_json)}")
            except Exception:
                raise ValueError(f"Error creating evaluation {response.text}")
        return CreateEvaluationResponse.model_validate(response.json())

    @classmethod
    def post_evaluation_results(
        cls, evaluation_id: uuid.UUID, data: list[EvaluationResultDatapoint]
    ) -> requests.Response:
        body = {
            "evaluationId": str(evaluation_id),
            "points": [datapoint.to_dict() for datapoint in data],
        }
        response = requests.post(
            cls.__base_http_url + "/v1/evaluation-datapoints",
            data=json.dumps(body),
            headers=cls._headers(),
        )
        if response.status_code != 200:
            try:
                resp_json = response.json()
                raise ValueError(
                    f"Failed to send evaluation results. Response: {json.dumps(resp_json)}"
                )
            except Exception:
                raise ValueError(
                    f"Failed to send evaluation results. Error: {response.text}"
                )
        return response

    @classmethod
    def update_evaluation_status(
        cls, evaluation_id: str, status: str
    ) -> UpdateEvaluationResponse:
        """
        Updates the status of an evaluation. Returns the updated evaluation object.

        Args:
            evaluation_id (str): The ID of the evaluation to update.
            status (str): The status to set for the evaluation.

        Returns:
            UpdateEvaluationResponse: The updated evaluation response.

        Raises:
            ValueError: If the request fails.
        """
        body = {
            "status": status,
        }
        url = f"{cls.__base_http_url}/v1/evaluations/{evaluation_id}"

        response = requests.post(
            url,
            data=json.dumps(body),
            headers=cls._headers(),
        )
        if response.status_code != 200:
            raise ValueError(
                f"Failed to update evaluation status {evaluation_id}. "
                f"Response: {response.text}"
            )

        return UpdateEvaluationResponse.model_validate(response.json())

    @classmethod
    def _headers(cls):
        assert cls.__project_api_key is not None, "Project API key is not set"
        return {
            "Authorization": "Bearer " + cls.__project_api_key,
            "Content-Type": "application/json",
        }



from enum import Enum


class Instruments(Enum):
    # The list of libraries which will be autoinstrumented
    # if no specific instruments are provided to initialize()
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    COHERE = "cohere"
    PINECONE = "pinecone"
    CHROMA = "chroma"
    GOOGLE_GENERATIVEAI = "google_generativeai"
    LANGCHAIN = "langchain"
    MISTRAL = "mistral"
    OLLAMA = "ollama"
    LLAMA_INDEX = "llama_index"
    MILVUS = "milvus"
    TRANSFORMERS = "transformers"
    TOGETHER = "together"
    BEDROCK = "bedrock"
    REPLICATE = "replicate"
    VERTEXAI = "vertexai"
    WATSONX = "watsonx"
    WEAVIATE = "weaviate"
    ALEPHALPHA = "alephalpha"
    MARQO = "marqo"
    LANCEDB = "lancedb"

    # The following libraries will not be autoinstrumented unless
    # specified explicitly in the initialize() call.
    REDIS = "redis"
    REQUESTS = "requests"
    URLLIB3 = "urllib3"
    PYMYSQL = "pymysql"



SPAN_INPUT = "lmnr.span.input"
SPAN_OUTPUT = "lmnr.span.output"
SPAN_TYPE = "lmnr.span.type"

ASSOCIATION_PROPERTIES = "lmnr.association.properties"
SESSION_ID = "session_id"
USER_ID = "user_id"
TRACE_TYPE = "trace_type"



# import pytest
# from openai import OpenAI
# from lmnr.traceloop_sdk.decorators import workflow


# @pytest.fixture
# def openai_client():
#     return OpenAI()


# @pytest.mark.vcr
# def test_resource_attributes(exporter, openai_client):
#     openai_client.chat.completions.create(
#         model="gpt-3.5-turbo",
#         messages=[{"role": "user", "content": "Tell me a joke about opentelemetry"}],
#     )

#     spans = exporter.get_finished_spans()
#     open_ai_span = spans[0]
#     assert open_ai_span.resource.attributes["something"] == "yes"
#     assert open_ai_span.resource.attributes["service.name"] == "test"


# def test_custom_span_processor(exporter_with_custom_span_processor):
#     @workflow()
#     def run_workflow():
#         pass

#     run_workflow()

#     spans = exporter_with_custom_span_processor.get_finished_spans()
#     workflow_span = spans[0]
#     assert workflow_span.attributes["custom_span"] == "yes"


# def test_instruments(exporter_with_custom_instrumentations):
#     @workflow()
#     def run_workflow():
#         pass

#     run_workflow()

#     spans = exporter_with_custom_instrumentations.get_finished_spans()
#     workflow_span = spans[0]
#     assert workflow_span


# def test_no_metrics(exporter_with_no_metrics):
#     @workflow()
#     def run_workflow():
#         pass

#     run_workflow()

#     spans = exporter_with_no_metrics.get_finished_spans()
#     workflow_span = spans[0]
#     assert workflow_span



import asyncio
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Awaitable, Optional, Set, Union
import uuid

from tqdm import tqdm

from ..traceloop_sdk.instruments import Instruments
from ..traceloop_sdk.tracing.attributes import SPAN_TYPE

from .laminar import Laminar as L
from .types import (
    CreateEvaluationResponse,
    Datapoint,
    EvaluationResultDatapoint,
    EvaluatorFunction,
    ExecutorFunction,
    Numeric,
    NumericTypes,
    SpanType,
    TraceType,
)
from .utils import is_async

DEFAULT_BATCH_SIZE = 5

_evaluation = None
_set_global_evaluation = False


@contextmanager
def set_global_evaluation(set_global_evaluation: bool):
    global _set_global_evaluation
    original = _set_global_evaluation
    try:
        _set_global_evaluation = set_global_evaluation
        yield
    finally:
        _set_global_evaluation = original
        pass


def get_evaluation_url(project_id: str, evaluation_id: str):
    return f"https://www.lmnr.ai/project/{project_id}/evaluations/{evaluation_id}"


class EvaluationReporter:
    def __init__(self):
        pass

    def start(self, name: str, project_id: str, id: str, length: int):
        print(f"Running evaluation {name}...\n")
        print(f"Check progress and results at {get_evaluation_url(project_id, id)}\n")
        self.cli_progress = tqdm(
            total=length,
            bar_format="{bar} {percentage:3.0f}% | ETA: {remaining}s | {n_fmt}/{total_fmt}",
            ncols=60,
        )

    def update(self, batch_length: int):
        self.cli_progress.update(batch_length)

    def stopWithError(self, error: Exception):
        self.cli_progress.close()
        sys.stderr.write(f"\nError: {error}\n")

    def stop(self, average_scores: dict[str, Numeric]):
        self.cli_progress.close()
        print("\nAverage scores:")
        for name, score in average_scores.items():
            print(f"{name}: {score}")
        print("\n")


class EvaluationDataset(ABC):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, idx) -> Datapoint:
        pass

    def slice(self, start: int, end: int):
        return [self[i] for i in range(max(start, 0), min(end, len(self)))]


class Evaluation:
    def __init__(
        self,
        data: Union[EvaluationDataset, list[Union[Datapoint, dict]]],
        executor: Any,
        evaluators: dict[str, EvaluatorFunction],
        name: Optional[str] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
        project_api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        http_port: Optional[int] = None,
        grpc_port: Optional[int] = None,
        instruments: Optional[Set[Instruments]] = None,
    ):
        """
        Initializes an instance of the Evaluations class.

        Parameters:
            data (Union[List[Union[EvaluationDatapoint, dict]], EvaluationDataset]): List of data points to evaluate or an evaluation dataset.
                            `data` is the input to the executor function,
                            `target` is the input to the evaluator function.
            executor (Callable[..., Any]): The executor function.
                            Takes the data point + any additional arguments
                            and returns the output to evaluate.
            evaluators (List[Callable[..., Any]]): List of evaluator functions.
                Each evaluator function takes the output of the executor _and_
                the target data, and returns a score. The score can be a
                single number or a record of string keys and number values.
                If the score is a single number, it will be named after the
                evaluator function. If the function is anonymous, it will be
                named `evaluator_${index}`, where index is the index of the
                evaluator function in the list starting from 1.
            name (Optional[str], optional): The name of the evaluation.
                            It will be auto-generated if not provided.
            batch_size (int, optional): The batch size for evaluation.
                            Defaults to DEFAULT_BATCH_SIZE.
            project_api_key (Optional[str], optional): The project API key.
                            Defaults to an empty string.
            base_url (Optional[str], optional): The base URL for the Laminar API.
                            Useful if self-hosted elsewhere.
                            Defaults to "https://api.lmnr.ai".
            http_port (Optional[int], optional): The port for the Laminar API HTTP service.
                            Defaults to 443.
            instruments (Optional[Set[Instruments]], optional): Set of modules to auto-instrument.
                            Defaults to None. If None, all available instruments will be used.
        """

        self.is_finished = False
        self.name = name
        self.reporter = EvaluationReporter()
        self.executor = executor
        self.evaluators = evaluators
        if isinstance(data, list):
            self.data = [
                (Datapoint.model_validate(point) if isinstance(point, dict) else point)
                for point in data
            ]
        else:
            self.data = data
        self.batch_size = batch_size
        L.initialize(
            project_api_key=project_api_key,
            base_url=base_url,
            http_port=http_port,
            grpc_port=grpc_port,
            instruments=instruments,
        )

    def run(self) -> Union[None, Awaitable[None]]:
        """Runs the evaluation.

        Creates a new evaluation if no evaluation with such name exists, or
        adds data to an existing one otherwise. Evaluates data points in
        batches of `self.batch_size`. The executor
        function is called on each data point to get the output,
        and then evaluate it by each evaluator function.

        Usage:
        ```python
        # in a synchronous context:
        e.run()
        # in an asynchronous context:
        await e.run()
        ```

        """
        if self.is_finished:
            raise Exception("Evaluation is already finished")

        loop = asyncio.get_event_loop()
        if loop.is_running():
            return loop.create_task(self._run())
        else:
            return loop.run_until_complete(self._run())

    async def _run(self) -> None:
        evaluation = L.create_evaluation(self.name)
        self.reporter.start(
            evaluation.name,
            evaluation.projectId,
            evaluation.id,
            len(self.data),
        )

        try:
            await self.evaluate_in_batches(evaluation)
        except Exception as e:
            L.update_evaluation_status(evaluation.id, "Error")
            self.reporter.stopWithError(e)
            self.is_finished = True
            return

        # If we update with status "Finished", we expect averageScores to be not empty
        updated_evaluation = L.update_evaluation_status(evaluation.id, "Finished")
        self.reporter.stop(updated_evaluation.averageScores)
        self.is_finished = True

    async def evaluate_in_batches(self, evaluation: CreateEvaluationResponse):
        for i in range(0, len(self.data), self.batch_size):
            batch = (
                self.data[i : i + self.batch_size]
                if isinstance(self.data, list)
                else self.data.slice(i, i + self.batch_size)
            )
            try:
                results = await self._evaluate_batch(batch)
                L.post_evaluation_results(evaluation.id, results)
            except Exception as e:
                print(f"Error evaluating batch: {e}")
            finally:
                self.reporter.update(len(batch))

    async def _evaluate_batch(
        self, batch: list[Datapoint]
    ) -> list[EvaluationResultDatapoint]:
        batch_promises = [self._evaluate_datapoint(datapoint) for datapoint in batch]
        results = await asyncio.gather(*batch_promises)
        return results

    async def _evaluate_datapoint(
        self, datapoint: Datapoint
    ) -> EvaluationResultDatapoint:
        with L.start_as_current_span("evaluation") as evaluation_span:
            L._set_trace_type(trace_type=TraceType.EVALUATION)
            evaluation_span.set_attribute(SPAN_TYPE, SpanType.EVALUATION.value)
            with L.start_as_current_span(
                "executor", input={"data": datapoint.data}
            ) as executor_span:
                executor_span.set_attribute(SPAN_TYPE, SpanType.EXECUTOR.value)
                output = (
                    await self.executor(datapoint.data)
                    if is_async(self.executor)
                    else self.executor(datapoint.data)
                )
                L.set_span_output(output)
            target = datapoint.target

            # Iterate over evaluators
            scores: dict[str, Numeric] = {}
            for evaluator_name, evaluator in self.evaluators.items():
                with L.start_as_current_span(
                    "evaluator", input={"output": output, "target": target}
                ) as evaluator_span:
                    evaluator_span.set_attribute(SPAN_TYPE, SpanType.EVALUATOR.value)
                    value = (
                        await evaluator(output, target)
                        if is_async(evaluator)
                        else evaluator(output, target)
                    )
                    L.set_span_output(value)

                # If evaluator returns a single number, use evaluator name as key
                if isinstance(value, NumericTypes):
                    scores[evaluator_name] = value
                else:
                    scores.update(value)

            trace_id = uuid.UUID(int=evaluation_span.get_span_context().trace_id)
            return EvaluationResultDatapoint(
                data=datapoint.data,
                target=target,
                executor_output=output,
                scores=scores,
                trace_id=trace_id,
            )


def evaluate(
    data: Union[EvaluationDataset, list[Union[Datapoint, dict]]],
    executor: ExecutorFunction,
    evaluators: dict[str, EvaluatorFunction],
    name: Optional[str] = None,
    batch_size: int = DEFAULT_BATCH_SIZE,
    project_api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    http_port: Optional[int] = None,
    grpc_port: Optional[int] = None,
    instruments: Optional[Set[Instruments]] = None,
) -> Optional[Awaitable[None]]:
    """
    If added to the file which is called through lmnr eval command, then simply registers the evaluation.
    Otherwise, if there is no event loop, creates it and runs the evaluation until completion.
    If there is an event loop, schedules the evaluation as a task in the event loop and returns an awaitable handle.

    Parameters:
        data (Union[List[Union[EvaluationDatapoint, dict]], EvaluationDataset]): List of data points to evaluate or an evaluation dataset.
                        `data` is the input to the executor function,
                        `target` is the input to the evaluator function.
        executor (Callable[..., Any]): The executor function.
                        Takes the data point + any additional arguments
                        and returns the output to evaluate.
        evaluators (List[Callable[..., Any]]): List of evaluator functions.
            Each evaluator function takes the output of the executor _and_
            the target data, and returns a score. The score can be a
            single number or a record of string keys and number values.
            If the score is a single number, it will be named after the
            evaluator function. If the function is anonymous, it will be
            named `evaluator_${index}`, where index is the index of the
            evaluator function in the list starting from 1.
        name (Optional[str], optional): The name of the evaluation.
            It will be auto-generated if not provided.
        batch_size (int, optional): The batch size for evaluation.
                        Defaults to DEFAULT_BATCH_SIZE.
        project_api_key (Optional[str], optional): The project API key.
                        Defaults to an empty string.
        base_url (Optional[str], optional): The base URL for the Laminar API.
                        Useful if self-hosted elsewhere.
                        Defaults to "https://api.lmnr.ai".
        http_port (Optional[int], optional): The port for the Laminar API HTTP service.
                        Defaults to 443.
        grpc_port (Optional[int], optional): The port for the Laminar API gRPC service.
                        Defaults to 8443.
        instruments (Optional[Set[Instruments]], optional): Set of modules to auto-instrument.
                        Defaults to None. If None, all available instruments will be used.
    """

    evaluation = Evaluation(
        data=data,
        executor=executor,
        evaluators=evaluators,
        name=name,
        batch_size=batch_size,
        project_api_key=project_api_key,
        base_url=base_url,
        http_port=http_port,
        grpc_port=grpc_port,
        instruments=instruments,
    )

    global _evaluation
    if _set_global_evaluation:
        _evaluation = evaluation
    else:
        return evaluation.run()



import logging


class CustomFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    green = "\x1b[32;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    fmt = "%(asctime)s::%(name)s::%(levelname)s: %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: grey + fmt + reset,
        logging.INFO: green + fmt + reset,
        logging.WARNING: yellow + fmt + reset,
        logging.ERROR: red + fmt + reset,
        logging.CRITICAL: bold_red + fmt + reset,
    }

    def format(self, record: logging.LogRecord):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# For StreamHandlers / console
class VerboseColorfulFormatter(CustomFormatter):
    def format(self, record):
        return super().format(record)


# For Verbose FileHandlers / files
class VerboseFormatter(CustomFormatter):
    fmt = "%(asctime)s::%(name)s::%(levelname)s| %(message)s (%(filename)s:%(lineno)d)"

    def format(self, record):
        formatter = logging.Formatter(self.fmt)
        return formatter.format(record)



from argparse import ArgumentParser
import asyncio
import importlib
import os
import sys

from lmnr.sdk.evaluations import set_global_evaluation


# TODO: Refactor this code
async def run_evaluation(args):
    sys.path.insert(0, os.getcwd())

    with set_global_evaluation(True):
        file = os.path.abspath(args.file)

        spec = importlib.util.spec_from_file_location("run_eval", file)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        from lmnr.sdk.evaluations import _evaluation
        evaluation = _evaluation
        await evaluation.run()


def cli():
    parser = ArgumentParser(
        prog="lmnr",
        description="CLI for Laminar",
    )

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    parser_eval = subparsers.add_parser("eval", description="Run an evaluation")
    parser_eval.add_argument("file", help="A file containing the evaluation to run")
    parser_eval.set_defaults(func=run_evaluation)

    parsed = parser.parse_args()
    asyncio.run(parsed.func(parsed))
